#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## variables.py
##
"""
    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        NumVarImpl
        IntVarImpl
        BoolVarImpl
        NegBoolView
        NDVarArray

    =================
    List of functions
    =================
    .. autosummary::
        :nosignatures:

        BoolVar
        IntVar
        cparray

    ==================
    Module description
    ==================

    This module is used for defining single variables as well as numpy-arrays of variables. There are 2 different types of variables: boolean variables, integer variables.
    
    Boolean Variables
    -----------------

    Boolean variables a.k.a `BoolVar` are variables that have a very specific domain. They take either the value `True` or `False` (1 or 0 respectively).
    The syntax is as follows:

    .. code-block:: python

        BoolVar([shape])

    - *optional* **shape**: integer value larger than 0 or tuple of integer values.

    The following examples show how to create a boolean variable with 3 use cases:

    - the creation of a single (unit-sized or non-vector) boolean variable.
        .. code-block:: python

            # creation of a unit Boolean variable
            x = BoolVar()

    - the creation of a vector boolean variables. 

        .. code-block:: python

            # creation of a vector Boolean variables
            x = BoolVar(3)

            # note that using the python unpacking you can assign them
            # to intermediate variables. THis allows for fine-grained use of variables when
            # defining the constraints of the model
            e,x,a,m,p,l = BoolVar(5)

    - the creation of array/tensor of boolean variables. 
        .. code-block:: python

            # creation of an __array__ of Boolean variables where (3, 8, 7) reflects
            # the dimensions of the tensor, a matrix of multiple-dimensions.
            # In this case, we create an 3D-array of dimensions 3 x 8 x 7.
            array_vars = BoolVar((3, 8, 7))

    Integer Variables
    -----------------

    Integer variables are variables that are given a lower bound and an upper bound, correpsonding to the values that they can take.
    The syntax is as follows:

    .. code-block:: python

        IntVar(lb, ub [, shape])
    
    - **lb**: lower bound
    - **ub**: upper bound
    - *optional* **shape**: integer value larger than 0 or tuple of integer values

    The following examples showcase how to instantiate integer variable with 3 use cases similar to `BoolVar`:

    - Creation of a single (unit-sized or non-vector) integer variable with a given lower bound (**lb**) of 3 and upper bound (**ub**) 8. Variable `x` can thus take values 3, 4, 5, 6, 7, 8 (upper bound included!).

        .. code-block:: python

            # creation of a unit integer variable with lowerbound of 3 and upperbound of 8 
            x = IntVar(3, 8)

    - Creation of a vector integer variables with all having the same given lower bound and upper bound:

        .. code-block:: python

            # creation of a vector Boolean of 5 variables with lowerbound of 3 and upperbound of 8 
            vecx = IntVar(3, 8, 5)

            # Similar `BoolVar`'s python unpacking can assign multiple intermediate variables at once
            e,x,a,m,p,l = IntVar(3, 8, 5)

    - Creation of a 4D-array/tensor (of dimensions 100 x 100 x 100 x 100) of boolean variables.
        .. code-block:: python

            arrx = IntVar(3, 8, (100, 100, 100, 100))

    Array of Variables
    ------------------

    N-dimensional array of cp-variables. Indexing an array with a variable is not allowed by standard numpy arrays, but it is allowed by cpmpy-numpy arrays. 
    First convert your numpy array to a cpmpy-numpy array with the `cparray()` wrapper:

    .. code-block:: python

        # Transforming a given numpy-array **m** into a cparray

        marr = np.array([
            [1, 2, 3, 4],
            [4, 8, 13, 15]
        ])

        m = cparray(marr)

    ==============
    Module details
    ==============
"""

from .utils.exceptions import NullShapeError
import numpy as np
from .expressions import Expression, Operator, is_num

# Helpers for type checking
def is_int(arg):
    return isinstance(arg, (int, np.integer))

def is_var(x):
    return isinstance(x, IntVarImpl)

class NumVarImpl(Expression):
    """
    Abstract **continuous numerical** variable with given lowerbound and upperbound.
    """
    def __init__(self, lb, ub, name):
        assert (is_num(lb) and is_num(ub))
        assert (lb <= ub)
        self.lb = lb
        self.ub = ub
        self.name = name
        self._value = None

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return False

    def value(self):
        return self._value
    
    def __repr__(self):
        return self.name

    # for sets/dicts. Because names are unique, so is the str repr
    def __hash__(self):
        return hash(str(self))

class IntVarImpl(NumVarImpl):
    """
    **Integer** constraint variable with given lowerbound and upperbound.
    """
    counter = 0

    def __init__(self, lb, ub, name=None):
        assert is_int(lb), "IntVar lowerbound must be integer {} {}".format(type(lb),lb)
        assert is_int(ub), "IntVar upperbound must be integer {} {}".format(type(ub),ub)

        if name is None:
            name = "IV{}".format(IntVarImpl.counter)
            IntVarImpl.counter = IntVarImpl.counter + 1 # static counter

        super().__init__(int(lb), int(ub), name=name) # explicit cast: can be numpy

class BoolVarImpl(IntVarImpl):
    """
    **Boolean** constraint variable with given lowerbound and upperbound.
    """
    counter = 0

    def __init__(self, lb=0, ub=1, name=None):
        assert(lb == 0 or lb == 1)
        assert(ub == 0 or ub == 1)

        if name is None:
            name = "BV{}".format(BoolVarImpl.counter)
            BoolVarImpl.counter = BoolVarImpl.counter + 1 # static counter
        IntVarImpl.__init__(self, lb, ub, name=name)
        

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return True

    def __invert__(self):
        return NegBoolView(self)

    def __eq__(self, other):
        # (BV == 1) <-> BV
        # if other == 1: XXX: dangerous because "=="" is overloaded 
        if other is 1 or other is True:
            return self
        if other is 0 or other is False:
            return ~self
        return super().__eq__(other)

    # when redefining __eq__, must redefine custom__hash__
    # https://stackoverflow.com/questions/53518981/inheritance-hash-sets-to-none-in-a-subclass
    def __hash__(self): return super().__hash__()

class NegBoolView(BoolVarImpl):
    """
        Represents not(`var`), not an actual variable implementation!

        It stores a link to `var`'s BoolVarImpl
    """
    def __init__(self, bv):
        #assert(isinstance(bv, BoolVarImpl))
        self._bv = bv

    def value(self):
        return not self._bv.value()

    def __repr__(self):
        return "~{}".format(self._bv.name)

    def __invert__(self):
        return self._bv


# subclass numericexpression for operators (first), ndarray for all the rest
class NDVarArray(Expression, np.ndarray):
    """
    N-dimensional numpy array of variables.
    """
    def __init__(self, shape, **kwargs):
        # TODO: global name?
        # this is nice and sneaky, 'self' is the list_of_arguments!
        Expression.__init__(self, "NDVarArray", self)
        # somehow, no need to call ndarray constructor

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return False

    def value(self):
        return np.reshape([x.value() for x in self], self.shape)
    
    def __getitem__(self, index):
        from .globalconstraints import Element # here to avoid circular
        # array access, check if variables are used in the indexing

        # index is single expression: direct element
        if isinstance(index, Expression):
            return Element([self, index])

        # index is array/tuple with at least one expr in it:
        # index non-expr part, and create element on expr part
        if isinstance(index, tuple) and \
           any(isinstance(el, Expression) for el in index):
            index_rest = list(index) # mutable view
            var = [] # collector of variables
            for i in range(len(index)):
                if isinstance(index[i], Expression):
                    index_rest[i] = Ellipsis # selects all remaining dimensions
                    var.append(index[i])
            assert (len(var)==1), "variable indexing (element) only supported with 1 variable at this moment"
            # single var, so flatten rest array
            array_rest = self[tuple(index_rest)] # non-var array selection
            return Element([array_rest, var[0]])

        ret = super().__getitem__(index)
        # this is a bit ugly,
        # but np.int and np.bool do not play well with > overloading
        if isinstance(ret, np.integer):
            return int(ret)
        elif isinstance(ret, np.bool_):
            return bool(ret)
        return ret

    def sum(self, axis=None, out=None):
        """
            overwrite np.sum(NDVarArray) as people might use it

            does not actually support axis/out... todo?
        """
        if not axis is None or not out is None:
            raise NotImplementedError() # please report on github with usecase

        # return sum object
        return Operator("sum", self)

    # TODO?
    #in	  __contains__(self, value) 	Check membership
    #object.__matmul__(self, other)


# N-dimensional array of Boolean Decision Variables
def BoolVar(shape=None, name=None):
    """
    # N-dimensional array of Boolean Decision Variables
    """
    if shape is None or shape == 1:
        return BoolVarImpl(name=name)
    elif shape == 0:
        raise NullShapeError(shape)
    
    # create base data
    data = np.array([BoolVarImpl(name=_genname(name, idxs)) for idxs in np.ndindex(shape)]) # repeat new instances
    # insert into custom ndarray
    return NDVarArray(shape, dtype=object, buffer=data)


def IntVar(lb, ub, shape=None, name=None):
    """
    N-dimensional array of Integer Decision Variables with lower-bound `lb` and upper-bound `ub`
    """
    if shape is None or shape == 1:
        return IntVarImpl(lb,ub, name=name)
    elif shape == 0:
        raise NullShapeError(shape)

    # create base data
    data = np.array([IntVarImpl(lb,ub, name=_genname(name, idxs)) for idxs in np.ndindex(shape)]) # repeat new instances
    # insert into custom ndarray
    return NDVarArray(shape, dtype=object, buffer=data)


def cparray(arr):
    """
    N-dimensional wrapper, wraps a standard array.

    So that we can do [1,2,3,4][var1] == var2, e.g. element([1,2,3,4],var1,var2)
    needed because standard arrays can not be indexed by non-constants
    """
    if not isinstance(arr, np.ndarray):
        arr = np.array(arr)
    return NDVarArray(shape=arr.shape, dtype=type(arr.flat[0]), buffer=arr)


def _genname(basename, idxs):
    """
    Helper function to 'name' array variables
    - idxs: list of indices, one for every dimension of the array
    - basename: base name to prepend

    if basename is 'None', then it returns None

    output: something like "basename[0,1]"
    """
    if basename == None:
        return None
    stridxs = ",".join(map(str,idxs))
    return f"{basename}[{stridxs}]" # "<name>[<idx0>,<idx1>,...]"
    



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## globalconstraints.py
##
"""
    ================
    List of functions
    =================

    .. autosummary::
        :nosignatures:

        min
        max

    ===============
    List of classes
    ===============

    .. autosummary::
        :nosignatures:

        alldifferent
        allequal
        circuit
        Table
        Minimum
        Maximum
        Element
        GlobalConstraint

    ==================
    Module description
    ==================

    Global constraint definitions

    A global constraint is nothing special in CPMpy. It is just an
    expression of type `GlobalConstraint` with a name and arguments.


    You can define a new global constraint as simply as:

    .. code-block:: python

        def my_global(args):
            return GlobalConstraint("my_global", args)


    Of course, solvers may not support a global constraint
    (if it does, it should be mapped to the API call in its SolverInterface)

    You can provide a decomposition for your global constraint through
    the decompose() function.
    To overwrite it, you should define your global constraint as a
    subclass of GlobalConstraint, rather then as a function above.

    Your decomposition function can use any standard CPMpy expression.
    For example:

    .. code-block:: python

        class my_global(GlobalConstraint):
            def __init__(self, args):
                super().__init__("my_global", args)

            def decompose(self):
                return [self.args[0] != self.args[1]] # your decomposition

    If you are modeling a problem and you want to use another decomposition,
    simply overwrite the 'decompose' function of the class, e.g.:

    .. code-block:: python

        def my_circuit_decomp(self):
            return [self.args[0] == 1] # does not actually enforce circuit
        circuit.decompose = my_circuit_decomp # attach it, no brackets!

        vars = IntVars(1,9, shape=(10,))
        constr = circuit(vars)

        Model(constr).solve()

    The above will use 'my_circuit_decomp', if the solver does not
    natively support 'circuit'.
"""
from .variables import *
from .expressions import *
from itertools import chain, combinations


class GlobalConstraint(Expression):
    # is_bool: whether this is normal constraint (True or False)
    #   not is_bool: it computes a numeric value (ex: Element)
    def __init__(self, name, arg_list, is_bool=True):
        super().__init__(name, arg_list)
        self._is_bool = is_bool

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return self._is_bool

    def decompose(self):
        """
            if a global constraint has a default decomposition,
            then it should monkey-patch this function, e.g.:
            def my_decomp_function(self):
                return []
            g = GlobalConstraint("g", args)
            g.decompose = my_decom_function
        """
        return None


# min: listwise 'min'
def min(iterable):
    """
        min() overwrites python built-in,
        checks if all constants and computes np.min() in that case
    """
    if not any(isinstance(elem, Expression) for elem in iterable):
        return np.min(iterable)
    return Minimum(iterable)

def max(iterable):
    """
        max() overwrites python built-in,
        checks if all constants and computes np.map() in that case
    """
    if not any(isinstance(elem, Expression) for elem in iterable):
        return np.max(iterable)
    return Maximum(iterable)



class alldifferent(GlobalConstraint):
    """
    all arguments have a different (distinct) value
    """
    def __init__(self, args):
        super().__init__("alldifferent", args)

    def decompose(self):
        return [var1 != var2 for var1, var2 in _all_pairs(self.args)]


class allequal(GlobalConstraint):
    """
    all arguments have the same value
    """
    def __init__(self, args):
        super().__init__("allequal", args)

    def decompose(self):
        return [var1 == var2 for var1, var2 in _all_pairs(self.args)]

class circuit(GlobalConstraint):
    """
    Variables of the constraint form a circuit ex: 0 -> 3 -> 2 -> 0
    """
    def __init__(self, args):
        super().__init__("circuit", args)

    def decompose(self):
        """
            TODO needs explanation/reference
        """
        n = len(self.args)
        a = self.args
        z = IntVar(0, n-1, n)
        constraints = [alldifferent(z),
                       alldifferent(a),
                       z[0]==a[0],
                       z[n-1]==0]
        for i in range(1,n-1):
            constraints += [z[i] != 0,
                            z[i] == a[z[i-1]]]
        return constraints

class Table(GlobalConstraint):
    """
    the values of the variables in 'array'
    correspond to a row in 'table'
    """
    def __init__(self, array, table):
        super().__init__("table", [array, table])

    def decompose(self):
        raise NotImplementedError("TODO: table decomposition")


class Minimum(GlobalConstraint):
    """
        Computes the minimum value of the arguments

        It is a 'functional' global constraint which implicitly returns a numeric variable
    """
    def __init__(self, arg_list):
        super().__init__("min", arg_list, is_bool=False)

    def value(self):
        return min([_argval(a) for a in self.args])

class Maximum(GlobalConstraint):
    """
        Computes the maximum value of the arguments

        It is a 'functional' global constraint which implicitly returns a numeric variable
    """
    def __init__(self, arg_list):
        super().__init__("max", arg_list, is_bool=False)

    def value(self):
        return max([_argval(a) for a in self.args])

class Element(GlobalConstraint):
    """
        The 'Element' global constraint enforces that the result equals Arr[Idx]
        with 'Arr' an array of constants of variables (the first argument)
        and 'Idx' an integer decision variable, representing the index into the array.

        Solvers implement it as Arr[Idx] == Y, but CPMpy will automatically derive or create
        an appropriate Y. Hence, you can write expressions like Arr[Idx] + 3 <= Y

        Element is a CPMpy built-in global constraint, so the class implements a few more
        extra things for convenience (.value() and .__repr__()). It is also an example of
        a 'numeric' global constraint.
    """

    def __init__(self, arg_list):
        assert (len(arg_list) == 2), "Element expression takes 2 arguments: Arr, Idx"
        super().__init__("element", arg_list, is_bool=False)

    def value(self):
        idxval = _argval(self.args[1])
        if not idxval is None:
            return _argval(self.args[0][idxval])
        return None # default

    def __repr__(self):
        return "{}[{}]".format(self.args[0], self.args[1])



def _all_pairs(args):
    """ internal helper function
    """
    pairs = list(combinations(args, 2))
    return pairs

# XXX, make argval shared util function?
def _argval(a):
    """ returns .value() of Expression, otherwise the variable itself
    """
    return a.value() if isinstance(a, Expression) else a



#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## expressions.py
##
"""
    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        Expression
        Comparison
        Operator

"""
import numpy as np

# Helpers for type checking
def is_num(arg):
    return isinstance(arg, (int, np.integer, float, np.float64))
def is_any_list(arg):
    return isinstance(arg, (list, tuple, np.ndarray))
def is_pure_list(arg):
    return isinstance(arg, (list, tuple))

# Overwriting all/any python built-ins
# all: listwise 'and'
def all(iterable):
    collect = [] # logical expressions
    for elem in iterable:
        if elem is False:
            return False # no need to create constraint
        elif elem is True:
            pass
        elif isinstance(elem, Expression):
            collect.append( elem.boolexpr() )
        else:
            raise Exception("unknown argument '{}' to 'all'".format(elem))
    if len(collect) == 1:
        return collect[0]
    if len(collect) >= 2:
        return Operator("and", collect)
    return True

# any: listwise 'or'
def any(iterable):
    collect = [] # logical expressions
    for elem in iterable:
        if elem is True:
            return True # no need to create constraint
        elif elem is False:
            pass
        elif isinstance(elem, Expression):
            collect.append( elem.boolexpr() )
        else:
            raise Exception("unknown argument '{}' to 'any'".format(elem))
    if len(collect) == 1:
        return collect[0]
    if len(collect) >= 2:
        return Operator("or", collect)
    return False


class Expression(object):
    """
    each Expression is a function with a self.name and self.args (arguments)
    each Expression is considered to be a function whose value can be used
      in other expressions
    each Expression may implement:
    - boolexpr(): the Boolean form of the expression
        default: (expr == 1)
        override for Boolean expressions (preferably through __eq__, see Comparison)
    - value(): the value of the expression, default None
    """

    def __init__(self, name, arg_list):
        if isinstance(arg_list, np.ndarray):
            # must flatten
            arg_list = arg_list.reshape(-1)
        for i in range(len(arg_list)):
            if isinstance(arg_list[i], np.ndarray):
                # must flatten
                arg_list[i] = arg_list[i].reshape(-1)

        assert (is_any_list(arg_list)), "_list_ of arguments required, even if of length one e.g. [arg]"
        self.name = name
        self.args = arg_list

    def __repr__(self):
        strargs = []
        for arg in self.args:
            if isinstance(arg, np.ndarray):
                # flatten
                strarg = ",".join(map(str,arg.flat))
                strargs.append( f"[{strarg}]" )
            else:
                strargs.append( f"{arg}" )
        return "{}({})".format(self.name, ",".join(strargs))

    # booleanised expression
    # optional, default: (self == 1)
    def boolexpr(self):
        return (self == 1)

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
            Default: yes
        """
        return True

    def value(self):
        return None # default

    # implication constraint: self -> other
    # Python does not offer relevant syntax...
    # for double implication, use equivalence self == other
    def implies(self, other):
        # other constant
        if other is True:
            return True
        if other is False:
            return ~self
        return Operator('->', [self, other])

    # Comparisons
    def __eq__(self, other):
        return Comparison("==", self, other)
    def __ne__(self, other):
        return Comparison("!=", self, other)
    def __lt__(self, other):
        return Comparison("<", self, other)
    def __le__(self, other):
        return Comparison("<=", self, other)
    def __gt__(self, other):
        return Comparison(">", self, other)
    def __ge__(self, other):
        return Comparison(">=", self, other)

    # Boolean Operators
    # Implements bitwise operations & | ^ and ~ (and, or, xor, not)
    def __and__(self, other):
        # some simple constant removal
        if other is True:
            return self
        if other is False:
            return False

        return Operator("and", [self, other])
    def __rand__(self, other):
        # some simple constant removal
        if other is True:
            return self
        if other is False:
            return False

        return Operator("and", [other, self])

    def __or__(self, other):
        # some simple constant removal
        if other is True:
            return True
        if other is False:
            return self

        return Operator("or", [self, other])
    def __ror__(self, other):
        # some simple constant removal
        if other is True:
            return True
        if other is False:
            return self

        return Operator("or", [other, self])

    def __xor__(self, other):
        # some simple constant removal
        if other is True:
            return ~self
        if other is False:
            return self

        return Operator("xor", [self, other])
    def __rxor__(self, other):
        # some simple constant removal
        if other is True:
            return ~self
        if other is False:
            return self

        return Operator("xor", [other, self])

    # Mathematical Operators, including 'r'everse if it exists
    # Addition
    def __add__(self, other):
        if is_num(other) and other == 0:
            return self
        return Operator("sum", [self, other])
    def __radd__(self, other):
        if is_num(other) and other == 0:
            return self
        return Operator("sum", [other, self])

    # substraction
    def __sub__(self, other):
        # if is_num(other) and other == 0:
        #     return self
        # return Operator("sub", [self, other])
        return self.__add__(-other)
    def __rsub__(self, other):
        # if is_num(other) and other == 0:
        #     return -self
        # return Operator("sub", [other, self])
        return (-self).__radd__(other)
    
    # multiplication, puts the 'constant' (other) first
    def __mul__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("mul", [self, other])
    def __rmul__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("mul", [other, self])

    # matrix multipliciation TODO?
    #object.__matmul__(self, other)

    # other mathematical ones
    def __truediv__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("div", [self, other])
    def __rtruediv__(self, other):
        return Operator("div", [other, self])

    def __mod__(self, other):
        if is_num(other) and other == 1:
            return self
        return Operator("mod", [self, other])
    def __rmod__(self, other):
        return Operator("mod", [other, self])

    def __pow__(self, other, modulo=None):
        assert (modulo is None), "Power operator: module not supported"
        return Operator("pow", [self, other])
    def __rpow__(self, other, modulo=None):
        assert (modulo is None), "Power operator: module not supported"
        return Operator("pow", [other, self])

    # Not implemented: (yet?)
    #object.__floordiv__(self, other)
    #object.__divmod__(self, other)

    # unary mathematical operators
    def __neg__(self):
        return Operator("-", [self])
    def __pos__(self):
        return self
    def __abs__(self):
        return Operator("abs", [self])
    # 'not' for now, no unary constraint for it but like boolexpr()
    def __invert__(self):
        return (self == 0)


class Comparison(Expression):
    allowed = {'==', '!=', '<=', '<', '>=', '>'}

    def __init__(self, name, left, right):
        assert (name in Comparison.allowed), "Symbol not allowed"
        # if vectorized, must match
        if hasattr(left, '__len__') and hasattr(right, '__len__'): 
            assert (len(left) == len(right)), "Comparison: arguments must have equal length"
        super().__init__(name, [left, right])

    def __repr__(self):
        if all(isinstance(x, Expression) for x in self.args):
            return "({}) {} ({})".format(self.args[0], self.name, self.args[1]) 
        # if not: prettier printing without braces
        return "{} {} {}".format(self.args[0], self.name, self.args[1]) 

    # it could be a vectorized constraint
    def __iter__(self):
        return (Comparison(self.name,l,r) for l,r in zip(self.args[0], self.args[1]))

    # is bool, check special case
    def __eq__(self, other):
        if is_num(other) and other == 1:
            return self
        return super().__eq__(other)
        
    # return the value of the expression
    # optional, default: None
    def value(self):
        arg_vals = [arg.value() if isinstance(arg, Expression) else arg for arg in self.args]
        if any(a is None for a in arg_vals): return None
        if   self.name == "==": return (arg_vals[0] == arg_vals[1])
        elif self.name == "!=": return (arg_vals[0] != arg_vals[1])
        elif self.name == "<":  return (arg_vals[0] < arg_vals[1])
        elif self.name == "<=": return (arg_vals[0] <= arg_vals[1])
        elif self.name == ">":  return (arg_vals[0] > arg_vals[1])
        elif self.name == ">=": return (arg_vals[0] >= arg_vals[1])
        return None # default


class Operator(Expression):
    """
    All kinds of operators on expressions,
    including mathematical and logical
    # convention for 2-ary operators: if one of the two is a constant,
    # it is stored first (as expr[0]), this eases weighted sum detection
    """
    allowed = {
        #name: (arity, is_bool)       arity 0 = n-ary, min 2
        'and': (0, True),
        'or':  (0, True),
        'xor': (0, True),
        '->':  (2, True),
        'sum': (0, False),
        'sub': (2, False), # x - y
        'mul': (2, False),
        'div': (2, False),
        'mod': (2, False),
        'pow': (2, False),
        '-':   (1, False), # -x
        'abs': (1, False),
    }
    printmap = {'sum': '+', 'sub': '-', 'mul': '*', 'div': '/'}

    def __init__(self, name, arg_list):
        # sanity checks
        assert (name in Operator.allowed), "Operator {} not allowed".format(name)
        arity, is_bool = Operator.allowed[name]
        if arity == 0:
            assert (len(arg_list) >= 2), "Operator: n-ary operators require at least two arguments"
        else:
            assert (len(arg_list) == arity), "Operator: {}, number of arguments must be {}".format(name, arity)
        if arity == 2 and hasattr(arg_list[0], '__len__'):
            # special case: can be two lists
            assert (len(arg_list[0]) == len(arg_list[1])), "Operator: the two arguments must have equal size"

        # convention for commutative binary operators:
        # swap if right is constant and left is not
        if len(arg_list) == 2 and is_num(arg_list[1]) and \
           name in {'sum', 'mul', 'and', 'or', 'xor'}:
            arg_list[0], arg_list[1] = arg_list[1], arg_list[0]

        super().__init__(name, arg_list)

    def is_bool(self):
        """ is it a Boolean (return type) Operator?
        """
        return Operator.allowed[self.name][1]
    
    def __repr__(self):
        printname = self.name
        if printname in Operator.printmap:
            printname = Operator.printmap[printname]

        # special cases
        if self.name == '-': # unary -
            return "-({})".format(self.args[0])

        # infix printing of two arguments
        if len(self.args) == 2:
            # bracketed printing of non-constants
            def wrap_bracket(arg):
                if isinstance(arg, Expression):
                    return f"({arg})"
                return arg
            return "{} {} {}".format(wrap_bracket(self.args[0]),
                                     printname,
                                     wrap_bracket(self.args[1]))

        return "{}({})".format(self.name, self.args)

    # it could be a vectorized constraint
    def __iter__(self):
        if len(self.args) == 2:
            return (Operator(self.name, list(args)) for args in zip(self.args[0], self.args[1]))
        return super().__iter__(self)

    # associative operations {'and', 'or', 'xor', 'sum', 'mul'} are chained
    # + special case for weighted sum (sum of mul)

    def __and__(self, other):
        if self.name == 'and':
            self.args.append(other)
            return self
        return super().__and__(other)
    def __rand__(self, other):
        if self.name == 'and':
            self.args.insert(0,other)
            return self
        return super().__rand__(other)

    def __or__(self, other):
        if self.name == 'or':
            self.args.append(other)
            return self
        return super().__or__(other)
    def __ror__(self, other):
        if self.name == 'or':
            self.args.insert(0,other)
            return self
        return super().__ror__(other)

    def __xor__(self, other):
        if self.name == 'xor':
            self.args.append(other)
            return self
        return super().__xor__(other)
    def __rxor__(self, other):
        if self.name == 'xor':
            self.args.insert(0,other)
            return self
        return super().__rxor__(other)

    def __add__(self, other):
        if is_num(other) and other == 0:
            return self

        if self.name == 'sum':
            self.args.append(other)
            return self
        return super().__add__(other)

    def __radd__(self, other):
        # only for constants
        if is_num(other) and other == 0:
            return self

        if self.name == 'sum':
            self.args.insert(0, other)
            return self
        return super().__radd__(other)

    # substraction; in case of sum or wsum, add
    def __sub__(self, other):
        if is_num(other) and other == 0:
            return self

        if self.name == 'sum':
            self.args.append(-other)
            return self
        return super().__sub__(other)
    def __rsub__(self, other):
        if is_num(other) and other == 0:
            return -self

        return (-self).__radd__(other)

    # is bool, check special case
    def __eq__(self, other):
        if is_num(other) and other == 1:
            # check if bool operator, do not add == 1
            arity, is_bool = Operator.allowed[self.name]
            if is_bool:
                return self
        return super().__eq__(other)

    def value(self):
        arg_vals = [arg.value() if isinstance(arg, Expression) else arg for arg in self.args]
        if any(a is None for a in arg_vals): return None
        if   self.name == "sum": return sum(arg_vals)
        elif self.name == "mul": return arg_vals[0] * arg_vals[1]
        elif self.name == "sub": return arg_vals[0] - arg_vals[1]
        elif self.name == "div": return arg_vals[0] / arg_vals[1]
        elif self.name == "mod": return arg_vals[0] % arg_vals[1]
        elif self.name == "pow": return arg_vals[0] ** arg_vals[1]
        elif self.name == "-":   return -arg_vals[0]
        elif self.name == "abs": return -arg_vals[0] if arg_vals[0] < 0 else arg_vals[0]
        return None # default




#!/usr/bin/env python
#-*- coding:utf-8 -*-
##
## model.py
##
"""
    ===============
    List of classes
    ===============
    .. autosummary::
        :nosignatures:

        Model

    ==================
    Module description
    ==================

    Class that contains the constraints and the objective function,
    and that provides an easy solve() abstraction which will call a solver.

    See the examples for basic usage, which involves:

    - creation, e.g. m = Model(cons, minimize=obj)
    - solving, e.g. m.solve()
    - optionally, checking status/runtime, e.g. m.status()

    ==============
    Module details
    ==============
"""
import numpy as np
from .expressions import Operator
from .solver_interfaces.util import get_supported_solvers
from .solver_interfaces.solver_interface import SolverInterface, SolverStatus, ExitStatus


class Model(object):
    """
    CPMpy Model object, contains the constraint and objective expression trees

    Arguments of constructor:

    - `*args`: Expression object(s) or list(s) of Expression objects
    - `minimize`: Expression object representing the objective to minimize
    - `maximize`: Expression object representing the objective to maximize

    At most one of minimize/maximize can be set, if none are set, it is assumed to be a satisfaction problem
    """
    def __init__(self, *args, minimize=None, maximize=None):
        assert ((minimize is None) or (maximize is None)), "can not set both minimize and maximize"
        self.cpm_status = SolverStatus("Model") # status of solving this model, will be replaced

        # list of constraints (arguments of top-level conjunction)
        if len(args) == 0 or (len(args) == 1 and isinstance(args[0], list) and len(args[0]) == 0): # None or empty list
            self.constraints = []
        else:
            root_constr = self._make_and_from_list(args)
            if root_constr.name == 'and':
                # unpack top-level conjuction
                self.constraints = root_constr.args
            else:
                # wrap in list
                self.constraints = [root_constr]

        # an expresion or None
        self.objective = None
        self.objective_max = None

        if not maximize is None:
            self.objective = maximize
            self.objective_max = True
        if not minimize is None:
            self.objective = minimize
            self.objective_max = False
        
    def __add__(self, cons):
        """
            Add one or more constraints to the model

            m = Model()
            m += [x > 0]
        """
        self.constraints += cons
        return self

    def __repr__(self):
        cons_str = ""
        for c in self.constraints:
            cons_str += "    {}\n".format(c)

        obj_str = ""
        if not self.objective is None:
            if self.objective_max:
                obj_str = "maximize "
            else:
                obj_str = "minimize "
        obj_str += str(self.objective)
            
        return "Constraints:\n{}Objective: {}".format(cons_str, obj_str)
    
    # solver: name of supported solver or any SolverInterface object
    def solve(self, solver=None, time_limit=None):
        """ Send the model to a solver and get the result

        :param solver: solver to use (verifies that the solver is supported on the current system)
        :type solver: None (default) or in [s.name in get_supported_solvers()] or a SolverInterface class (Class, not object! e.g. CPMpyOrTools, not CPMpyOrTools()!)

        :param time_limit: optional, time limit in seconds
        :type time_limit: int or float

        :return: the computed output:
            - True      if it is a satisfaction problem and it is satisfiable
            - False     if it is a satisfaction problem and not satisfiable
            - [int]     if it is an optimisation problem
        """
        solver_class = None
        # get supported solvers
        supsolvers = get_supported_solvers()
        if solver is None: # default is first
            solver_class = supsolvers[0]
        elif not isinstance(solver, SolverInterface):
            solvername = solver
            for s in supsolvers:
                if s.name == solvername:
                    solver_class = s
                    break # break and hence 'solver' is correct object
        elif isinstance(solver, SolverInterface) and solver.supported():
            solver_class = solver
        else:
            raise Exception("'{}' is not in the list of supported solvers and not a SolverInterface class".format(solver))
        assert(solver_class is not None)
                
        # instatiate solver with this model
        s = solver_class(self)
        # call solver
        ret = s.solve(time_limit=time_limit)
        # store CPMpy status (s object has no further use)
        self.cpm_status = s.status()
        return ret

    def status(self):
        """
            Returns the status of the latest solver run on this model

            Status information includes exit status (optimality) and runtime.

        :return: an object of :class:`SolverStatus`
        """
        return self.cpm_status

    def _make_and_from_list(self, args):
        """ recursively reads a list of Expression and returns the 'And' conjunctive of the elements in the list """
        lst = list(args) # make mutable copy of type list
        # do recursive where needed, with overwrite
        for (i, expr) in enumerate(lst):
            if isinstance(expr, list):
                lst[i] = self._make_and_from_list(expr)
        if len(lst) == 1:
            return lst[0]
        return Operator("and", lst)



class NullShapeError(Exception):
    def __init__(self, shape, message="Shape should be non-zero"):
        self.shape = shape
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        return f'{self.shape}: {self.message}'